__all__ = ["model"]

from icevision.imports import *
from icevision.backbones import resnet_fpn
from icevision.models.torchvision.utils import *

from torchvision.models.detection.faster_rcnn import (
    fasterrcnn_resnet50_fpn,
    FasterRCNN,
    FastRCNNPredictor,
)


def model(
    num_classes: int,
    backbone: Optional[nn.Module] = None,
    remove_internal_transforms: bool = True,
    pretrained: bool = True,
    **faster_rcnn_kwargs
) -> nn.Module:
    """FasterRCNN model implemented by torchvision.

    # Arguments
        num_classes: Number of classes.
        backbone: Backbone model to use. Defaults to a resnet50_fpn model.
        remove_internal_transforms: The torchvision model internally applies transforms
        like resizing and normalization, but we already do this at the `Dataset` level,
        so it's safe to remove those internal transforms.
        pretrained: Argument passed to `fastercnn_resnet50_fpn` if `backbone is None`.
        By default it is set to True: this is generally used when training a new model (transfer learning).
        `pretrained = False`  is used during inference (prediction) for cases where the users have their own pretrained weights.
        **faster_rcnn_kwargs: Keyword arguments that internally are going to be passed to
        `torchvision.models.detection.faster_rcnn.FastRCNN`.

    # Returns
        A Pytorch `nn.Module`.
    """
    if backbone is None:
        model = fasterrcnn_resnet50_fpn(
            pretrained=pretrained, pretrained_backbone=pretrained, **faster_rcnn_kwargs
        )

        in_features = model.roi_heads.box_predictor.cls_score.in_features
        model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes)

        resnet_fpn.patch_param_groups(model.backbone)
    else:
        model = FasterRCNN(backbone, num_classes=num_classes, **faster_rcnn_kwargs)

    patch_rcnn_param_groups(model=model)

    if remove_internal_transforms:
        remove_internal_model_transforms(model)

    return model



__all__ = ["open_img", "show_img", "plot_grid", "plot_grid_preds_actuals"]

from icevision.imports import *


def open_img(fn, gray=False):
    if not os.path.exists(fn):
        raise ValueError(f"File {fn} does not exists")
    color = cv2.COLOR_BGR2GRAY if gray else cv2.COLOR_BGR2RGB
    return cv2.cvtColor(cv2.imread(str(fn), cv2.IMREAD_UNCHANGED), color)


def show_img(img, ax=None, show: bool = False, **kwargs):
    img = img.squeeze().copy()
    cmap = "gray" if len(img.shape) == 2 else None

    if ax is None:
        fig, ax = plt.subplots(**kwargs)

    ax.imshow(img, cmap=cmap)
    ax.set_axis_off()

    if show:
        plt.show()

    return ax


def plot_grid(fs: List[callable], ncols=1, figsize=None, show=False, **kwargs):
    nrows = math.ceil(len(fs) / ncols)
    figsize = figsize or (12 * ncols, 12 * nrows)

    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, **kwargs)
    axs = np.asarray(axs)
    for f, ax in zip(fs, axs.flatten()):
        f(ax=ax)

    plt.tight_layout()
    if show:
        plt.show()


def plot_grid_preds_actuals(
    actuals, predictions, figsize=None, show=False, annotations=None, **kwargs
):
    fig, axs = plt.subplots(
        nrows=len(actuals),
        ncols=2,
        figsize=figsize or (6, 6 * len(actuals) / 2 / 0.75),
        **kwargs,
    )
    i = 0
    for im, ax in zip(zip(actuals, predictions), axs.reshape(-1, 2)):
        ax[0].imshow(im[0], cmap=None)
        ax[0].set_title("Ground truth")
        ax[1].imshow(im[1], cmap=None)
        ax[1].set_title("Prediction")

        if annotations is None:
            ax[0].set_axis_off()
            ax[1].set_axis_off()
        else:
            ax[0].get_xaxis().set_ticks([])
            ax[0].set_frame_on(False)
            ax[0].get_yaxis().set_visible(False)
            ax[0].set_xlabel(annotations[i][0], ma="left")

            ax[1].get_xaxis().set_ticks([])
            ax[1].set_frame_on(False)
            ax[1].get_yaxis().set_visible(False)
            ax[1].set_xlabel(annotations[i][1], ma="left")

            i += 1

    plt.tight_layout()
    if show:
        plt.show()
    return axs



from icevision.backbones.vgg import *
from icevision.backbones.resnet import *
from icevision.backbones.mobilenet import *

import icevision.backbones.resnet_fpn
import icevision.backbones.resnest_fpn



from icevision.models.mmdet.common.mask.dataloaders import *
from icevision.models.mmdet.common.mask.prediction import *
from icevision.models.mmdet.common.mask.show_results import *

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    from icevision.models.mmdet import fastai

if SoftDependencies.pytorch_lightning:
    from icevision.models.mmdet import lightning



__all__ = ["voc", "VocXmlParser", "VocMaskParser"]

import xml.etree.ElementTree as ET
from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.parsers.parser import *
from icevision.parsers.defaults import *
from icevision.parsers.mixins import *


def voc(
    annotations_dir: Union[str, Path],
    images_dir: Union[str, Path],
    class_map: Optional[ClassMap] = None,
    masks_dir: Optional[Union[str, Path]] = None,
):
    if not masks_dir:
        return VocXmlParser(
            annotations_dir=annotations_dir,
            images_dir=images_dir,
            class_map=class_map,
        )
    else:
        return VocMaskParser(
            annotations_dir=annotations_dir,
            images_dir=images_dir,
            masks_dir=masks_dir,
            class_map=class_map,
        )


class VocXmlParser(Parser, FilepathMixin, SizeMixin, LabelsMixin, BBoxesMixin):
    def __init__(
        self,
        annotations_dir: Union[str, Path],
        images_dir: Union[str, Path],
        class_map: Optional[ClassMap] = None,
    ):
        super().__init__(class_map=class_map)
        self.images_dir = Path(images_dir)

        self.annotations_dir = Path(annotations_dir)
        self.annotation_files = get_files(self.annotations_dir, extensions=[".xml"])

    def __len__(self):
        return len(self.annotation_files)

    def __iter__(self):
        yield from self.annotation_files

    def prepare(self, o):
        tree = ET.parse(str(o))
        self._root = tree.getroot()
        self._filename = self._root.find("filename").text
        self._size = self._root.find("size")

    def imageid(self, o) -> Hashable:
        return str(Path(self._filename).stem)

    def filepath(self, o) -> Union[str, Path]:
        return self.images_dir / self._filename

    def image_width_height(self, o) -> Tuple[int, int]:
        return get_image_size(self.filepath(o))

    def labels(self, o) -> List[Hashable]:
        labels = []
        for object in self._root.iter("object"):
            label = object.find("name").text
            labels.append(label)

        return labels

    def bboxes(self, o) -> List[BBox]:
        def to_int(x):
            return int(float(x))

        bboxes = []
        for object in self._root.iter("object"):
            xml_bbox = object.find("bndbox")
            xmin = to_int(xml_bbox.find("xmin").text)
            ymin = to_int(xml_bbox.find("ymin").text)
            xmax = to_int(xml_bbox.find("xmax").text)
            ymax = to_int(xml_bbox.find("ymax").text)

            bbox = BBox.from_xyxy(xmin, ymin, xmax, ymax)
            bboxes.append(bbox)

        return bboxes


class VocMaskParser(VocXmlParser, MasksMixin):
    def __init__(
        self,
        annotations_dir: Union[str, Path],
        images_dir: Union[str, Path],
        masks_dir: Union[str, Path],
        class_map: Optional[ClassMap] = None,
    ):
        super().__init__(
            annotations_dir=annotations_dir, images_dir=images_dir, class_map=class_map
        )
        self.masks_dir = masks_dir
        self.mask_files = get_image_files(masks_dir)

        self._imageid2maskfile = {self.imageid_mask(o): o for o in self.mask_files}

        # filter annotations
        masks_ids = frozenset(self._imageid2maskfile.keys())
        self._intersection = []
        for item in super().__iter__():
            super().prepare(item)
            if super().imageid(item) in masks_ids:
                self._intersection.append(item)

    def __len__(self):
        return len(self._intersection)

    def __iter__(self):
        yield from self._intersection

    def imageid_mask(self, o) -> Hashable:
        """Should return the same as `imageid` from parent parser."""
        return str(Path(o).stem)

    def masks(self, o) -> List[Mask]:
        mask_file = self._imageid2maskfile[self.imageid(o)]
        return [VocMaskFile(mask_file)]



__all__ = ["BBox"]

from icevision.imports import *
from .exceptions import *


class BBox:
    """Bounding Box representation.

    Should **not** be instantiated directly, instead use `from_*` methods.
    e.g. `from_xyxy`, `from_xywh`.
    Is able to transform coordinates into different formats,
    e.g. `xyxy`, `xywh`.

    # Examples

    Create from `xywh` format, and get `xyxy` coordinates.
    ```python
    bbox = BBox.from_xywh(1, 1, 4, 4)
    xyxy = bbox.xyxy
    ```
    """

    def __init__(self, xmin, ymin, xmax, ymax):
        self.xmin, self.ymin, self.xmax, self.ymax = xmin, ymin, xmax, ymax

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} "
            f"(xmin:{self.xmin}, ymin:{self.ymin}, xmax:{self.xmax}, ymax:{self.ymax})>"
        )

    def __eq__(self, other) -> bool:
        if isinstance(other, BBox):
            return self.xyxy == other.xyxy
        return False

    @property
    def width(self):
        return self.xmax - self.xmin

    @property
    def height(self):
        return self.ymax - self.ymin

    @property
    def area(self):
        return self.width * self.height

    def to_tensor(self):
        return tensor(self.xyxy, dtype=torch.float)

    def autofix(self, img_w, img_h) -> bool:
        """Tries to automatically fix invalid coordinates.

        # Returns
        - False if nothing was fixed (data had no problems)
        - True if data was successfully fixed
        - Raises `InvalidDataError` if unables to automatically fix the data
        """
        # conditions where data can be fixed
        if self.xmin < 0:
            logger.log(
                "AUTOFIX-SUCCESS",
                f"Clipping bbox xmin from {self.xmin} to 0 (Before: {self})",
            )
            self.xmin = max(self.xmin, 0)

        if self.ymin < 0:
            logger.log(
                "AUTOFIX-SUCCESS",
                f"Clipping bbox ymin from {self.ymin} to 0 (Before: ({self}))",
            )
            self.ymin = max(self.ymin, 0)

        if self.xmax > img_w:
            logger.log(
                "AUTOFIX-SUCCESS",
                f"Clipping bbox xmax from {self.xmax} to image width {img_w} (Before: {self})",
            )
            self.xmax = min(self.xmax, img_w)

        if self.ymax > img_h:
            logger.log(
                "AUTOFIX-SUCCESS",
                f"Clipping bbox ymax from {self.ymax} to image height {img_h} (Before: {self})",
            )
            self.ymax = min(self.ymax, img_h)

        # conditions where data cannot be fixed
        if (self.xmin >= self.xmax) or (self.ymin >= self.ymax):
            msg = []
            if self.xmin >= self.xmax:
                msg += [
                    f"\tx_min:{self.xmin} is greater than or equal to x_max:{self.xmax}"
                ]
            if self.ymin >= self.ymax:
                msg += [
                    f"\ty_min:{self.ymin} is greater than or equal to y_max:{self.ymax}"
                ]

            msg = "\n".join(msg)
            raise InvalidDataError(f"Cannot auto-fix coordinates: {self}\n{msg}")

        if self.xmin < 0 or self.ymin < 0 or self.xmax > img_w or self.ymax > img_h:
            return True

        return False

    @property
    def xyxy(self):
        return (self.xmin, self.ymin, self.xmax, self.ymax)

    @property
    def yxyx(self):
        return (self.ymin, self.xmin, self.ymax, self.xmax)

    @property
    def xywh(self):
        return (self.xmin, self.ymin, self.width, self.height)

    def relative_xcycwh(self, img_width: int, img_height: int):
        scale = np.array([img_width, img_height, img_width, img_height])
        x, y, w, h = self.xywh / scale
        xc = x + 0.5 * w
        yc = y + 0.5 * h
        return (xc, yc, w, h)

    @classmethod
    def from_xywh(cls, x, y, w, h):
        return cls(x, y, x + w, y + h)

    @classmethod
    def from_xyxy(cls, xl, yu, xr, yb):
        return cls(xl, yu, xr, yb)

    @classmethod
    def from_relative_xcycwh(cls, xc, yc, bw, bh, img_width, img_height):
        # subtracting 0.5 goes from center to left/upper edge, adding goes to right/bottom
        pnts = [(xc - 0.5 * bw), (yc - 0.5 * bh), (xc + 0.5 * bw), (yc + 0.5 * bh)]
        # convert from relative to absolute coordinates
        scale = np.array([img_width, img_height, img_width, img_height])
        xl, yu, xr, yb = np.around(pnts * scale).astype(int).tolist()
        return cls.from_xyxy(xl, yu, xr, yb)

    @classmethod
    def from_rle(cls, rle, h, w):
        a = np.array(rle.counts, dtype=int)
        a = a.reshape((-1, 2))  # an array of (start, length) pairs
        a[:, 0] -= 1  # `start` is 1-indexed
        y0 = a[:, 0] % h
        y1 = y0 + a[:, 1]
        if np.any(y1 > h):
            # got `y` overrun, meaning that there are a pixels in mask on 0 and shape[0] position
            y0 = 0
            y1 = h
        else:
            y0 = np.min(y0)
            y1 = np.max(y1)
        x0 = a[:, 0] // h
        x1 = (a[:, 0] + a[:, 1]) // h
        x0 = np.min(x0)
        x1 = np.max(x1)
        if x1 > w:
            # just went out of the image dimensions
            raise ValueError(f"invalid RLE or image dimensions: x1={x1} > shape[1]={w}")
        return cls.from_xyxy(x0, y0, x1, y1)



from icevision.models.utils import *
from icevision.models.interpretation import *

# backwards compatibility
from icevision.models.torchvision import (
    faster_rcnn,
    mask_rcnn,
    retinanet,
    keypoint_rcnn,
)
from icevision.models import torchvision


# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.effdet:
    # backwards compatibility
    from icevision.models.ross import efficientdet
    from icevision.models import ross

if SoftDependencies.mmdet:
    from icevision.models import mmdet



__all__ = ["BaseRecord", "autofix_records", "create_mixed_record"]

from icevision.imports import *
from icevision.utils import *
from collections.abc import MutableMapping
from copy import copy
from .record_mixins import *
from .exceptions import *


# TODO: MutableMapping because of backwards compatability
class BaseRecord(ImageidRecordMixin, SizeRecordMixin, RecordMixin, MutableMapping):
    def num_annotations(self) -> Dict[str, int]:
        return self._num_annotations()

    def check_num_annotations(self):
        num_annotations = self.num_annotations()
        if len(set(num_annotations.values())) > 1:
            msg = "\n".join([f"\t- {v} for {k}" for k, v in num_annotations.items()])
            raise AutofixAbort(
                "Number of items should be the same for each annotation type"
                f", but got:\n{msg}"
            )

    def autofix(self):
        self.check_num_annotations()

        success_dict = self._autofix()
        success_list = np.array(list(success_dict.values()))
        if len(success_list) == 0:
            return success_dict
        keep_mask = reduce(np.logical_and, success_list)
        discard_idxs = np.where(keep_mask == False)[0]

        for i in discard_idxs:
            logger.log(
                "AUTOFIX-REPORT",
                "Removed annotation with index: {}, "
                "for more info check the AUTOFIX-FAIL messages above",
                i,
            )
            self.remove_annotation(i)

        return success_dict

    def remove_annotation(self, i):
        # TODO: remove_annotation might work incorrectly with masks
        # TODO: fixed with EncodedRLEs?
        self._remove_annotation(i)

    def aggregate_objects(self):
        return self._aggregate_objects()

    def copy(self) -> "BaseRecord":
        return copy(self)

    def load(self) -> "BaseRecord":
        record = copy(self)
        record._load()
        return record

    def __repr__(self) -> str:
        _reprs = self._repr()
        _repr = "".join(f"\n\t- {o}" for o in _reprs)
        return f"Record:{_repr}"

    # backwards compatiblity: implemented method to behave like a dict
    def __getitem__(self, key):
        return self.as_dict()[key]

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        delattr(self, key)

    def __iter__(self):
        yield from self.as_dict()

    def __len__(self):
        return len(self.as_dict())


def autofix_records(records: Sequence[BaseRecord]) -> Sequence[BaseRecord]:
    keep_records = []
    for record in records:

        def _pre_replay():
            logger.log(
                "AUTOFIX-START",
                "️🔨  Autofixing record with imageid: {}  ️🔨",
                record.imageid,
            )

        with ReplaySink(_pre_replay) as sink:
            try:
                record.autofix()
                keep_records.append(record)
            except AutofixAbort as e:
                logger.warning(
                    "🚫 Record could not be autofixed and will be removed because: {}",
                    str(e),
                )

    return keep_records


def create_mixed_record(
    mixins: Sequence[Type[RecordMixin]], add_base: bool = True
) -> Type[BaseRecord]:
    mixins = (BaseRecord, *mixins) if add_base else tuple(mixins)

    TemporaryRecord = type("Record", mixins, {})
    class_name = "".join([o.__name__ for o in TemporaryRecord.mro()])

    Record = type(class_name, mixins, {})
    return patch_class_to_main(Record)



from icevision.models.ross.efficientdet.model import *
from icevision.models.ross.efficientdet.dataloaders import *
from icevision.models.ross.efficientdet.loss_fn import *
from icevision.models.ross.efficientdet.prediction import *
from icevision.models.ross.efficientdet.show_results import *

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    from icevision.models.ross.efficientdet import fastai

if SoftDependencies.pytorch_lightning:
    from icevision.models.ross.efficientdet import lightning



__all__ = ["MMDetectionCallback"]

from icevision.imports import *
from icevision.engines.fastai import *


class _ModelWrap(nn.Module):
    def __init__(self, model: nn.Module):
        super().__init__()
        self.model = model

    def forward(self, xb):
        return self.model.train_step(data=xb, optimizer=None)


class MMDetectionCallback(fastai.Callback):
    def after_create(self):
        self.learn.model = _ModelWrap(self.model)

    def before_batch(self):
        self.learn.records = self.yb[0]
        self.learn.yb = self.xb



__all__ = ["show_results", "interp"]

from icevision.models.torchvision.faster_rcnn.show_results import *
from icevision.models.interpretation import Interpretation
from icevision.models.torchvision.retinanet.dataloaders import (
    valid_dl,
    infer_dl,
)
from icevision.models.torchvision.retinanet.prediction import (
    predict_dl,
)

_LOSSES_DICT = {
    "classification": [],
    "bbox_regression": [],
    "loss_total": [],
}

interp = Interpretation(
    losses_dict=_LOSSES_DICT,
    valid_dl=valid_dl,
    infer_dl=infer_dl,
    predict_dl=predict_dl,
)



__all__ = ["model"]

from icevision.imports import *
from icevision.models.torchvision.utils import *
from icevision.backbones import resnet_fpn

from torchvision.models.detection.retinanet import (
    retinanet_resnet50_fpn,
    RetinaNet,
    RetinaNetHead,
)


def model(
    num_classes: int,
    backbone: Optional[nn.Module] = None,
    remove_internal_transforms: bool = True,
    pretrained: bool = True,
    **retinanet_kwargs
) -> nn.Module:
    if backbone is None:
        model = retinanet_resnet50_fpn(
            pretrained=pretrained, pretrained_backbone=pretrained, **retinanet_kwargs
        )
        model.head = RetinaNetHead(
            in_channels=model.backbone.out_channels,
            num_anchors=model.head.classification_head.num_anchors,
            num_classes=num_classes,
        )
        resnet_fpn.patch_param_groups(model.backbone)
    else:
        model = RetinaNet(
            backbone=backbone, num_classes=num_classes, **retinanet_kwargs
        )

    patch_retinanet_param_groups(model)

    if remove_internal_transforms:
        remove_internal_model_transforms(model)

    return model



__all__ = ["Adapter", "aug_tfms", "resize_and_pad"]

import albumentations as A
from itertools import chain
from icevision.imports import *
from icevision.core import *
from icevision.tfms.transform import *


def _resize(size, ratio_resize=A.LongestMaxSize):
    return ratio_resize(size) if isinstance(size, int) else A.Resize(*size)


def resize_and_pad(
    size: Union[int, Tuple[int, int]],
    pad: A.DualTransform = partial(
        A.PadIfNeeded, border_mode=cv2.BORDER_CONSTANT, value=[124, 116, 104]
    ),
):
    height, width = (size, size) if isinstance(size, int) else size
    return [_resize(size), pad(min_height=height, min_width=width)]


def aug_tfms(
    size: Union[int, Tuple[int, int]],
    presize: Optional[Union[int, Tuple[int, int]]] = None,
    horizontal_flip: Optional[A.HorizontalFlip] = A.HorizontalFlip(),
    shift_scale_rotate: Optional[A.ShiftScaleRotate] = A.ShiftScaleRotate(),
    rgb_shift: Optional[A.RGBShift] = A.RGBShift(),
    lightning: Optional[A.RandomBrightnessContrast] = A.RandomBrightnessContrast(),
    blur: Optional[A.Blur] = A.Blur(blur_limit=(1, 3)),
    crop_fn: Optional[A.DualTransform] = partial(A.RandomSizedBBoxSafeCrop, p=0.5),
    pad: Optional[A.DualTransform] = partial(
        A.PadIfNeeded, border_mode=cv2.BORDER_CONSTANT, value=[124, 116, 104]
    ),
) -> List[A.BasicTransform]:
    """Collection of useful augmentation transforms.

    # Arguments
        size: The final size of the image. If an `int` is given, the maximum size of
            the image is rescaled, maintaing aspect ratio. If a `tuple` is given,
            the image is rescaled to have that exact size (height, width).
        presizing: Rescale the image before applying other transfroms. If `None` this
                transform is not applied. First introduced by fastai,this technique is
                explained in their book in [this](https://github.com/fastai/fastbook/blob/master/05_pet_breeds.ipynb)
                chapter (tip: search for "Presizing").
        horizontal_flip: Flip around the y-axis. If `None` this transform is not applied.
        shift_scale_rotate: Randomly shift, scale, and rotate. If `None` this transform
                is not applied.
        rgb_shift: Randomly shift values for each channel of RGB image. If `None` this
                transform is not applied.
        lightning: Randomly changes Brightness and Contrast. If `None` this transform
                is not applied.
        blur: Randomly blur the image. If `None` this transform is not applied.
        crop_fn: Randomly crop the image. If `None` this transform is not applied.
                Use `partial` to saturate other parameters of the class.
        pad: Pad the image to `size`, squaring the image if `size` is an `int`.
            If `None` this transform is not applied. Use `partial` to sature other
            parameters of the class.

    # Returns
        A list of albumentations transforms.
    """

    height, width = (size, size) if isinstance(size, int) else size

    tfms = []
    tfms += [_resize(presize, A.SmallestMaxSize) if presize is not None else None]
    tfms += [horizontal_flip, shift_scale_rotate, rgb_shift, lightning, blur]
    # Resize as the last transforms to reduce the number of artificial artifacts created
    if crop_fn is not None:
        crop = crop_fn(height=height, width=width)
        tfms += [A.OneOrOther(crop, _resize(size), p=crop.p)]
    else:
        tfms += [_resize(size)]
    tfms += [pad(min_height=height, min_width=width) if pad is not None else None]

    tfms = [tfm for tfm in tfms if tfm is not None]

    return tfms


class Adapter(Transform):
    """Adapter that enables the use of albumentations transforms.

    # Arguments
        tfms: `Sequence` of albumentation transforms.
    """

    def __init__(self, tfms: Sequence[A.BasicTransform]):
        self.tfms_list = tfms
        self.bbox_params = A.BboxParams(format="pascal_voc", label_fields=["labels"])
        self.keypoint_params = A.KeypointParams(
            format="xy", remove_invisible=False, label_fields=["keypoints_labels"]
        )
        super().__init__(
            tfms=A.Compose(
                tfms, bbox_params=self.bbox_params, keypoint_params=self.keypoint_params
            )
        )

    def apply(
        self,
        img: np.ndarray,
        labels=None,
        bboxes: List[BBox] = None,
        masks: MaskArray = None,
        iscrowds: List[int] = None,
        keypoints: List[KeyPoints] = None,
        **kwargs
    ):
        # Substitue labels with list of idxs, so we can also filter out iscrowds in case any bboxes are removed
        # TODO: Same should be done if a masks is completely removed from the image (if bboxes is not given)
        params = {"image": img}
        params["labels"] = list(range(len(labels))) if labels is not None else []
        params["bboxes"] = [o.xyxy for o in bboxes] if bboxes is not None else []

        if keypoints is not None:
            flat_tfms_list_ = _flatten_tfms(self.tfms_list)
            if get_transform(flat_tfms_list_, "RandomSizedBBoxSafeCrop") is not None:
                raise RuntimeError(
                    "RandomSizedBBoxSafeCrop is not supported for keypoints"
                )

            k = [xy for o in keypoints for xy in o.xy]
            c = [label for o in keypoints for label in o.metadata.labels]
            v = [visible for o in keypoints for visible in o.visible]
            assert len(k) == len(c) == len(v)
            params["keypoints"] = k
            params["keypoints_labels"] = c

        if masks is not None:
            params["masks"] = list(masks.data)

        if bboxes is None:
            self.tfms.processors.pop("bboxes", None)
        if keypoints is None:
            self.tfms.processors.pop("keypoints", None)

        d = self.tfms(**params)

        img_h, img_w = _get_size_without_padding(self.tfms_list, img, d["image"])
        # We use the values in d['labels'] to get what was removed by the transform
        if labels is not None:
            keep_mask = np.zeros(len(labels), dtype=bool)
            keep_mask[d["labels"]] = True
        else:
            keep_mask = None

        out = {"img": d["image"]}
        out["height"], out["width"] = img_h, img_w

        if labels is not None:
            out["labels"] = _filter_attribute(labels, keep_mask)

        if bboxes is not None:
            # TODO: quickfix from 576
            # bb = [_clip_bboxes(xyxy, img_h, img_w) for xyxy in d["bboxes"]]
            bb = [xyxy for xyxy in d["bboxes"]]
            out["bboxes"] = [BBox.from_xyxy(*xyxy) for xyxy in bb]

        if masks is not None:
            keep_masks = _filter_attribute(d["masks"], keep_mask)
            out["masks"] = MaskArray(np.array(keep_masks))

        if keypoints is not None:
            # remove_invisible=False, therefore all points getting in are also getting out
            assert len(d["keypoints"]) == len(k)

            tfmed_kpts = _remove_outside_keypoints(d["keypoints"], img_h, img_w, v)
            # flatten list of keypoints
            flat_kpts = list(chain.from_iterable(tfmed_kpts))
            # group keypoints from same instance
            group_kpts = [
                flat_kpts[i : i + len(flat_kpts) // len(keypoints)]
                for i in range(0, len(flat_kpts), len(flat_kpts) // len(keypoints))
            ]
            assert len(group_kpts) == len(keypoints)

            kpts = [
                KeyPoints.from_xyv(group_kpt, original_kpt.metadata)
                for group_kpt, original_kpt in zip(group_kpts, keypoints)
            ]
            out["keypoints"] = _filter_attribute(kpts, keep_mask)

        if iscrowds is not None:
            out["iscrowds"] = _filter_attribute(iscrowds, keep_mask)

        return out


def _filter_attribute(v: list, keep_mask: Union[List[bool], None]):
    if keep_mask is None:
        return v
    assert len(v) == len(keep_mask)
    return [o for o, keep in zip(v, keep_mask) if keep]


def _remove_outside_keypoints(tfms_kps, h, w, v):
    """Remove keypoints that are outside image dimensions."""
    v_n = v.copy()
    tra_n = tfms_kps.copy()
    for i in range(len(tfms_kps)):
        if v[i] > 0:
            v_n[i] = _check_kps_coords(tfms_kps[i], h, w)
            if v_n[i] == 1:
                v_n[i] = v[i]
        if v_n[i] == 0:
            tra_n[i] = (0, 0)
        tra_n[i] = (tra_n[i][0], tra_n[i][1], v_n[i])
    return tra_n


def _clip_bboxes(xyxy, h, w):
    """Clip bboxes coordinates that are outside image dimensions."""
    x1, y1, x2, y2 = xyxy
    if w >= h:
        pad = (w - h) // 2
        h1 = pad
        h2 = w - pad
        return (x1, max(y1, h1), x2, min(y2, h2))
    else:
        pad = (h - w) // 2
        w1 = pad
        w2 = h - pad
        return (max(x1, w1), y1, min(x2, w2), y2)


def _get_size_without_padding(
    tfms_list, before_tfm_img, after_tfm_img
) -> Tuple[int, int]:
    height, width, _ = after_tfm_img.shape

    if get_transform(tfms_list, "Pad") is not None:
        after_pad_h, after_pad_w, _ = before_tfm_img.shape

        t = get_transform(tfms_list, "SmallestMaxSize")
        if t is not None:
            presize = t.max_size
            height, width = _func_max_size(after_pad_h, after_pad_w, presize, min)

        t = get_transform(tfms_list, "LongestMaxSize")
        if t is not None:
            size = t.max_size
            height, width = _func_max_size(after_pad_h, after_pad_w, size, max)

    return height, width


def py3round(number):
    """Unified rounding in all python versions."""
    if abs(round(number) - number) == 0.5:
        return int(2.0 * round(number / 2.0))

    return int(round(number))


def _func_max_size(height, width, max_size, func):
    scale = max_size / float(func(width, height))

    if scale != 1.0:
        height, width = tuple(py3round(dim * scale) for dim in (height, width))
    return height, width


def get_transform(tfms_list, t):
    for el in tfms_list:
        if t in str(type(el)):
            return el
    return None


def _check_kps_coords(p, h, w):
    x, y = p
    if w >= h:
        pad = (w - h) // 2
        h1 = pad
        h2 = w - pad
        return int((x <= w) and (x >= 0) and (y >= h1) and (y <= h2))
    else:
        pad = (h - w) // 2
        w1 = pad
        w2 = h - pad
        return int((x <= w2) and (x >= w1) and (y >= 0) and (y <= h))


def _is_iter(o):
    try:
        i = iter(o)
        return True
    except:
        return False


def _flatten_tfms(t):
    flat = []
    for o in t:
        if _is_iter(o):
            flat += [i for i in o]
        else:
            flat.append(o)
    return flat



__all__ = ["show_results", "interp"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.models.base_show_results import base_show_results
from icevision.models.interpretation import Interpretation
from icevision.models.torchvision.mask_rcnn.dataloaders import (
    build_infer_batch,
    valid_dl,
    infer_dl,
)
from icevision.models.torchvision.mask_rcnn.prediction import (
    predict,
    predict_dl,
)


def show_results(
    model: nn.Module,
    dataset: Dataset,
    class_map: Optional[ClassMap] = None,
    detection_threshold: float = 0.5,
    mask_threshold: float = 0.5,
    num_samples: int = 6,
    ncols: int = 3,
    denormalize_fn: Optional[callable] = denormalize_imagenet,
    show: bool = True,
    device: Optional[torch.device] = None,
) -> None:
    return base_show_results(
        predict_fn=predict,
        build_infer_batch_fn=build_infer_batch,
        model=model,
        dataset=dataset,
        class_map=class_map,
        num_samples=num_samples,
        ncols=ncols,
        denormalize_fn=denormalize_fn,
        show=show,
        detection_threshold=detection_threshold,
        mask_threshold=mask_threshold,
        device=device,
    )


_LOSSES_DICT = {
    "loss_classifier": [],
    "loss_box_reg": [],
    "loss_objectness": [],
    "loss_rpn_box_reg": [],
    "loss_mask": [],
    "loss_total": [],
}

interp = Interpretation(
    losses_dict=_LOSSES_DICT,
    valid_dl=valid_dl,
    infer_dl=infer_dl,
    predict_dl=predict_dl,
)



from icevision.models.torchvision import backbones
from icevision.models.torchvision.loss_fn import *

from icevision.models.torchvision.keypoint_rcnn.parser import *
from icevision.models.torchvision.keypoint_rcnn.dataloaders import *
from icevision.models.torchvision.keypoint_rcnn.model import *

from icevision.models.torchvision.keypoint_rcnn.prediction import *

from icevision.models.torchvision.keypoint_rcnn.show_results import *


# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    import icevision.models.torchvision.keypoint_rcnn.fastai

if SoftDependencies.pytorch_lightning:
    import icevision.models.torchvision.keypoint_rcnn.lightning



__all__ = ["Dataset"]

from icevision.imports import *
from icevision.core import *
from icevision.tfms import *


class Dataset:
    """Container for a list of records and transforms.

    Steps each time an item is requested (normally via directly indexing the `Dataset`):
        * Grab a record from the internal list of records.
        * Prepare the record (open the image, open the mask, add metadata).
        * Apply transforms to the record.

    # Arguments
        records: A list of records.
        tfm: Transforms to be applied to each item.
    """

    def __init__(
        self,
        records: List[dict],
        tfm: Transform = None,
    ):
        self.records = records
        self.tfm = tfm

    def __len__(self):
        return len(self.records)

    def __getitem__(self, i):
        data = self.records[i].load().as_dict()
        if self.tfm is not None:
            data = self.tfm(data)
        return data

    def __repr__(self):
        return f"<{self.__class__.__name__} with {len(self.records)} items>"

    @classmethod
    def from_images(cls, images: Sequence[np.array], tfm: Transform = None):
        """Creates a `Dataset` from a list of images.

        # Arguments
            images: `Sequence` of images in memory (numpy arrays).
            tfm: Transforms to be applied to each item.
        """
        Record = create_mixed_record((ImageRecordMixin,))
        records = []
        for i, image in enumerate(images):
            record = Record()
            record.set_imageid(i)
            record.set_img(image)
            records.append(record)

        return cls(records=records, tfm=tfm)



__all__ = [
    "Parser",
    "DetectionDefaultParser",
    "MaskDefaultParser",
    "KeyPointsDefaultParser",
]

from icevision.imports import *
from icevision.core import *
from icevision import parsers


class Parser(parsers.Parser):
    def __init__(self, class_map: Optional[ClassMap] = None):
        super().__init__(self, class_map=class_map)
        self.class_map.set_background(0)


class DetectionDefaultParser(
    Parser,
    parsers.FilepathMixin,
    parsers.LabelsMixin,
    parsers.BBoxesMixin,
):
    pass


class MaskDefaultParser(
    Parser,
    parsers.FilepathMixin,
    parsers.LabelsMixin,
    parsers.BBoxesMixin,
    parsers.MasksMixin,
    parsers.IsCrowdsMixin,  # TODO: is IsCrowds really necessary?
):
    pass


class KeyPointsDefaultParser(
    Parser,
    parsers.FilepathMixin,
    parsers.LabelsMixin,
    parsers.BBoxesMixin,
    parsers.KeyPointsMixin,
):
    pass



__all__ = ["ParserInterface", "Parser"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.parsers.mixins import *


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


# TODO: Rename to BaseParser
class ParserInterface(ABC):
    @abstractmethod
    def parse(
        self, data_splitter: DataSplitter, autofix: bool = True, show_pbar: bool = True
    ) -> List[List[RecordType]]:
        pass


class Parser(ClassMapMixin, ImageidMixin, SizeMixin, ParserInterface, ABC):
    """Base class for all parsers, implements the main parsing logic.

    The actual fields to be parsed are defined by the mixins used when
    defining a custom parser. The only required fields for all parsers
    are `image_id` and `image_width_height`.

    # Examples

    Create a parser for image filepaths.
    ```python
    class FilepathParser(Parser, FilepathParserMixin):
        # implement required abstract methods
    ```
    """

    def __init__(self, class_map: Optional[ClassMap] = None):
        self.class_map = class_map or ClassMap()
        if class_map is None:
            self.class_map.unlock()

    @abstractmethod
    def __iter__(self) -> Any:
        pass

    def prepare(self, o):
        pass

    def record_class(self) -> BaseRecord:
        return create_mixed_record(self.record_mixins())

    def parse_dicted(
        self, idmap: IDMap, show_pbar: bool = True
    ) -> Dict[int, RecordType]:

        Record = self.record_class()
        records = {}

        for sample in pbar(self, show_pbar):
            try:
                self.prepare(sample)
                true_imageid = self.imageid(sample)
                imageid = idmap[true_imageid]

                try:
                    record = records[imageid]
                except KeyError:
                    record = Record()

                self.parse_fields(sample, record)

                # HACK: fix imageid (needs to be transformed with idmap)
                record.set_imageid(imageid)
                records[imageid] = record

            except AbortParseRecord as e:
                logger.warning(
                    "Record with imageid: {} was skipped because: {}",
                    true_imageid,
                    str(e),
                )

        return dict(records)

    def _check_path(self, path: Union[str, Path] = None):
        if path is None:
            return False
        if path is not None:
            return Path(path).exists()

    def parse(
        self,
        data_splitter: DataSplitter = None,
        idmap: IDMap = None,
        autofix: bool = True,
        show_pbar: bool = True,
        cache_filepath: Union[str, Path] = None,
    ) -> List[List[BaseRecord]]:
        """Loops through all data points parsing the required fields.

        # Arguments
            data_splitter: How to split the parsed data, defaults to a [0.8, 0.2] random split.
            idmap: Maps from filenames to unique ids, pass an `IDMap()` if you need this information.
            show_pbar: Whether or not to show a progress bar while parsing the data.
            cache_filepath: Path to save records in pickle format. Defaults to None, e.g.
                            if the user does not specify a path, no saving nor loading happens.

        # Returns
            A list of records for each split defined by `data_splitter`.
        """

        # Hack to define the class of the mixed_record in the local namespace. This is required for pickeling.
        Record = self.record_class()

        if self._check_path(cache_filepath):
            logger.info(
                f"Loading cached records from {cache_filepath}",
            )
            return pickle.load(open(Path(cache_filepath), "rb"))
        else:
            idmap = idmap or IDMap()
            data_splitter = data_splitter or RandomSplitter([0.8, 0.2])
            records = self.parse_dicted(show_pbar=show_pbar, idmap=idmap)

            splits = data_splitter(idmap=idmap)
            all_splits_records = []
            if autofix:
                logger.opt(colors=True).info("<blue><bold>Autofixing records</></>")
            for ids in splits:
                split_records = [records[i] for i in ids if i in records]

                if autofix:
                    split_records = autofix_records(split_records)

                all_splits_records.append(split_records)

            self.class_map.lock()
            if cache_filepath is not None:
                pickle.dump(all_splits_records, open(Path(cache_filepath), "wb"))

            return all_splits_records

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return ["def __iter__(self) -> Any:"] + templates

    @classmethod
    def generate_template(cls):
        for template in cls._templates():
            print(f"{template}")



__all__ = ["model"]

from icevision.imports import *
from mmcv import Config
from mmdet.models import build_detector
from mmcv.runner import load_checkpoint


def model(
    cfg_path: Union[str, Path],
    num_classes: int,
    weights_path: Optional[Union[str, Path]] = None,
) -> nn.Module:
    cfg = Config.fromfile(str(cfg_path))
    cfg.model.roi_head.bbox_head.num_classes = num_classes - 1
    if weights_path is not None:
        cfg.model.pretrained = None

    _model = build_detector(cfg.model, cfg.train_cfg, cfg.test_cfg)

    if weights_path is not None:
        load_checkpoint(_model, str(weights_path))

    return _model



__all__ = [
    "ParserMixin",
    "ClassMapMixin",
    "ImageidMixin",
    "FilepathMixin",
    "SizeMixin",
    "LabelsMixin",
    "BBoxesMixin",
    "AreasMixin",
    "MasksMixin",
    "IsCrowdsMixin",
    "KeyPointsMixin",
]

from icevision.imports import *
from icevision.core import *
from icevision.utils import *


class ParserMixin(ABC):
    def record_mixins(self) -> List[RecordMixin]:
        return []

    def parse_fields(self, o, record) -> None:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        return []


class ClassMapMixin(ParserMixin):
    def record_mixins(self):
        return [ClassMapRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record):
        record.set_class_map(self.class_map)
        super().parse_fields(o, record)


class ImageidMixin(ParserMixin):
    """Adds `imageid` method to parser"""

    def record_mixins(self):
        return [ImageidRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record):
        record.set_imageid(self.imageid(o))
        super().parse_fields(o, record)

    @abstractmethod
    def imageid(self, o) -> Hashable:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        return ["def imageid(self, o) -> Hashable:"] + super()._templates()


class FilepathMixin(ParserMixin):
    """Adds `filepath` method to parser"""

    def record_mixins(self):
        return [FilepathRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record):
        filepath = Path(self.filepath(o))

        if not filepath.exists():
            raise AbortParseRecord(f"File '{filepath}' does not exist")

        record.set_filepath(filepath)
        super().parse_fields(o, record)

    @abstractmethod
    def filepath(self, o) -> Union[str, Path]:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        return ["def filepath(self, o) -> Union[str, Path]:"] + super()._templates()


class SizeMixin(ParserMixin):
    """Adds `image_width_height` method to parser"""

    def record_mixins(self):
        return [SizeRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record):
        # TODO: rename to image_weight, image_height
        width, height = self.image_width_height(o)
        record.set_image_size(width=width, height=height)
        super().parse_fields(o, record)

    @abstractmethod
    def image_width_height(self, o) -> Tuple[int, int]:
        """Returns the image size (width, height)."""

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        # Dynamic template based if FilepathMixin is present
        if issubclass(cls, FilepathMixin):
            return [
                "def image_width_height(self, o) -> Tuple[int, int]:\n"
                "    return get_image_size(self.filepath(o))",
            ] + templates
        return [
            "def image_width_height(self, o) -> Tuple[int, int]:",
        ] + templates


### Annotation parsers ###
class LabelsMixin(ParserMixin):
    """Adds `labels` method to parser"""

    def record_mixins(self) -> List[RecordMixin]:
        return [LabelsRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record):
        # super first because class_map need to be set before
        super().parse_fields(o, record)

        names = self.labels(o)
        ids = [self.class_map.get_name(name) for name in names]
        record.add_labels(ids)

    @abstractmethod
    def labels(self, o) -> List[Hashable]:
        """Returns the labels for the receive sample.

        !!! danger "Important"
        Return `BACKGROUND` (imported from icevision) for background.
        """

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def labels(self, o) -> List[Hashable]:"]


class BBoxesMixin(ParserMixin):
    """Adds `bboxes` method to parser"""

    def record_mixins(self) -> List[RecordMixin]:
        return [BBoxesRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record):
        record.add_bboxes(self.bboxes(o))
        super().parse_fields(o, record)

    @abstractmethod
    def bboxes(self, o) -> List[BBox]:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def bboxes(self, o) -> List[BBox]:"]


class MasksMixin(ParserMixin):
    """Adds `masks` method to parser"""

    def record_mixins(self) -> List[RecordMixin]:
        return [MasksRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record) -> None:
        super().parse_fields(o, record)
        record.add_masks(self.masks(o))

    @abstractmethod
    def masks(self, o) -> List[Mask]:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def masks(self, o) -> List[Mask]:"]


class AreasMixin(ParserMixin):
    """Adds `areas` method to parser"""

    def record_mixins(self) -> List[RecordMixin]:
        return [AreasRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record) -> None:
        record.add_areas(self.areas(o))
        super().parse_fields(o, record)

    @abstractmethod
    def areas(self, o) -> List[float]:
        """Returns areas of the segmentation"""

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def areas(self, o) -> List[float]:"]


class IsCrowdsMixin(ParserMixin):
    """Adds `iscrowds` method to parser"""

    def record_mixins(self) -> List[RecordMixin]:
        return [IsCrowdsRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record):
        record.add_iscrowds(self.iscrowds(o))
        super().parse_fields(o, record)

    @abstractmethod
    def iscrowds(self, o) -> List[bool]:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def iscrowds(self, o) -> List[bool]:"]


class KeyPointsMixin(ParserMixin):
    """Adds `keypoints` method to parser"""

    def record_mixins(self) -> List[RecordMixin]:
        return [KeyPointsRecordMixin, *super().record_mixins()]

    def parse_fields(self, o, record):
        record.add_keypoints(self.keypoints(o))
        super().parse_fields(o, record)

    @abstractmethod
    def keypoints(self, o) -> List[KeyPoints]:
        pass

    @classmethod
    def _templates(cls) -> List[str]:
        templates = super()._templates()
        return templates + ["def keypoints(self, o) -> List[KeyPoints]:"]



from icevision.models.torchvision.loss_fn import *

from icevision.models.torchvision.retinanet import backbones
from icevision.models.torchvision.retinanet.parser import *
from icevision.models.torchvision.retinanet.dataloaders import *
from icevision.models.torchvision.retinanet.model import *
from icevision.models.torchvision.retinanet.prediction import *
from icevision.models.torchvision.retinanet.show_results import *

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    import icevision.models.torchvision.retinanet.fastai

if SoftDependencies.pytorch_lightning:
    import icevision.models.torchvision.retinanet.lightning



# The draw functions are based from:
# https://github.com/fizyr/keras-retinanet/blob/master/keras_retinanet/utils/visualization.py
# https://github.com/fizyr/keras-maskrcnn/blob/master/keras_maskrcnn/utils/visualization.py

__all__ = [
    "draw_sample",
    "draw_record",
    "draw_pred",
    "draw_bbox",
    "draw_mask",
    "draw_keypoints",
]

from icevision.imports import *
from icevision.data import *
from icevision.core import *


def draw_sample(
    sample,
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    display_keypoints: bool = True,
):
    img = sample["img"].copy()
    if denormalize_fn is not None:
        img = denormalize_fn(img)

    for label, bbox, mask, keypoints in itertools.zip_longest(
        sample.get("labels", []),
        sample.get("bboxes", []),
        sample.get("masks", []),
        sample.get("keypoints", []),
    ):
        color = (np.random.random(3) * 0.6 + 0.4) * 255
        color = tuple(color.astype(int).tolist())

        if display_mask and mask is not None:
            img = draw_mask(img=img, mask=mask, color=color)
        if display_bbox and bbox is not None:
            img = draw_bbox(img=img, bbox=bbox, color=color)
        if display_label and label is not None:
            img = draw_label(
                img=img,
                label=label,
                bbox=bbox,
                mask=mask,
                class_map=class_map,
                color=color,
            )
        if display_keypoints and keypoints is not None:
            img = draw_keypoints(img=img, kps=keypoints, color=color)

    return img


def draw_label(
    img: np.ndarray,
    label: int,
    color,
    class_map: Optional[ClassMap] = None,
    bbox=None,
    mask=None,
):
    # finds label position based on bbox or mask
    if bbox is not None:
        x, y, _, _ = bbox.xyxy
    elif mask is not None:
        y, x = np.unravel_index(mask.data.argmax(), mask.data.shape)
    else:
        x, y = 0, 0

    if class_map is not None:
        caption = class_map.get_id(label)
    else:
        caption = str(label)

    return _draw_label(img=img, caption=caption, x=int(x), y=int(y), color=color)


def _draw_label(
    img: np.ndarray,
    caption: str,
    x: int,
    y: int,
    color,
    font=cv2.FONT_HERSHEY_SIMPLEX,
    font_scale: float = 1.0,
):
    """Draws a caption above the box in an image."""
    y -= 10
    w, h = cv2.getTextSize(caption, font, fontScale=font_scale, thickness=1)[0]

    # make the coords of the box with a small padding of two pixels
    # check if the box_pt2 is inside the image otherwise invert the label box (meaning the label box will be inside the bounding box)
    if (y - h - 2) > 0:
        box_pt1, box_pt2 = (x, y + 10), (x + w + 2, y - h - 2)
    else:
        box_pt1, box_pt2 = (x, y + h + 22), (x + w + 2, y + 10)

    cv2.rectangle(img, box_pt1, box_pt2, color, cv2.FILLED)

    label_pt = (box_pt1[0], box_pt1[1] - 10)
    cv2.putText(img, caption, label_pt, font, font_scale, (240, 240, 240), 2)

    return img


def draw_record(
    record,
    class_map: Optional[ClassMap] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
    display_keypoints: bool = True,
):
    sample = record.load()
    return draw_sample(
        sample=sample,
        class_map=class_map,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
        display_keypoints=display_keypoints,
    )


def draw_pred(
    img: np.ndarray,
    pred: dict,
    class_map: Optional[ClassMap] = None,
    denormalize_fn: Optional[callable] = None,
    display_label: bool = True,
    display_bbox: bool = True,
    display_mask: bool = True,
):
    sample = pred.copy()
    sample["img"] = img
    return draw_sample(
        sample=sample,
        class_map=class_map,
        denormalize_fn=denormalize_fn,
        display_label=display_label,
        display_bbox=display_bbox,
        display_mask=display_mask,
    )


def draw_bbox(
    img: np.ndarray,
    bbox: BBox,
    color: Tuple[int, int, int],
    gap: bool = True,
):
    """Draws a box on an image with a given color.
    # Arguments
        image     : The image to draw on.
        box       : A list of 4 elements (x1, y1, x2, y2).
        color     : The color of the box.
    """

    # Calculate image dimensions
    dims = sorted(img.shape, reverse=True)

    # corner thickness is linearly correlated with the smaller image dimension.
    # We use the smaller image dimension rather than image area so as to avoid
    # overly thick lines for large non-square images prior to transforming
    # images. We set lower and upper bounds for corner thickness.
    min_corner = 1
    max_corner = 15
    corner_thickness = int(0.005 * dims[1] + min_corner)
    if corner_thickness > max_corner:
        corner_thickness = int(max_corner)

    corner_length = int(0.021 * dims[1] + 2.25)

    # inner thickness of bboxes with corners
    inner_thickness = int(1 + 0.0005 * dims[1])

    # bbox thickness of bboxes without corners
    min_bbox = 1
    max_bbox = 8
    bbox_thickness = int(0.0041 * dims[1] - 0.0058)
    if bbox_thickness < min_bbox:
        bbox_thickness = min_bbox
    if bbox_thickness > max_bbox:
        bbox_thickness = int(max_bbox)

    if gap == False:
        xyxy = tuple(np.array(bbox.xyxy, dtype=int))
        cv2.rectangle(img, xyxy[:2], xyxy[2:], color, bbox_thickness, cv2.LINE_AA)
        return img

    xmin, ymin, xmax, ymax = tuple(np.array(bbox.xyxy, dtype=int))

    points = [0] * 12
    points[0] = (xmin, ymin + corner_length)
    points[1] = (xmin, ymin)
    points[2] = (xmin + corner_length, ymin)

    points[3] = (xmax - corner_length, ymin)
    points[4] = (xmax, ymin)
    points[5] = (xmax, ymin + corner_length)

    points[6] = (xmax, ymax - corner_length)
    points[7] = (xmax, ymax)
    points[8] = (xmax - corner_length, ymax)

    points[9] = (xmin + corner_length, ymax)
    points[10] = (xmin, ymax)
    points[11] = (xmin, ymax - corner_length)

    if (
        ymax - (ymin + 4 * corner_length) < corner_length
        or xmax - (xmin + 4 * corner_length) < corner_length
    ):
        for i in range(4):
            cv2.line(
                img,
                points[i * 3 + 1],
                points[10 - 3 * i],
                color,
                bbox_thickness,
                cv2.LINE_4,
            )
        for i in range(2):
            cv2.line(
                img,
                points[6 * i + 1],
                points[i * 6 + 4],
                color,
                bbox_thickness,
                cv2.LINE_4,
            )
    else:
        for i in range(2):
            for j in range(2):
                cv2.line(
                    img,
                    points[i * 6 + j * 4],
                    points[i * 6 + j * 4 + 1],
                    color,
                    corner_thickness,
                    cv2.LINE_AA,
                )
                cv2.line(
                    img,
                    points[i * 3 + 1],
                    points[10 - 3 * i],
                    color,
                    inner_thickness,
                    cv2.LINE_4,
                )
        for i in range(2):
            for j in range(2):
                cv2.line(
                    img,
                    points[i * 6 + j * 2 + 1],
                    points[i * 6 + j * 2 + 2],
                    color,
                    corner_thickness,
                    cv2.LINE_4,
                )
                cv2.line(
                    img,
                    points[6 * i + 1],
                    points[i * 6 + 4],
                    color,
                    inner_thickness,
                    cv2.LINE_AA,
                )

    return img


def draw_mask(
    img: np.ndarray, mask: MaskArray, color: Tuple[int, int, int], blend: float = 0.5
):
    color = np.asarray(color, dtype=int)
    # draw mask
    mask_idxs = np.where(mask.data)
    img[mask_idxs] = blend * img[mask_idxs] + (1 - blend) * color

    # draw border
    border = mask.data - cv2.erode(mask.data, np.ones((7, 7), np.uint8), iterations=1)
    border_idxs = np.where(border)
    img[border_idxs] = color

    return img


def draw_keypoints(
    img: np.ndarray,
    kps: KeyPoints,
    color: Tuple[int, int, int],
):
    x, y, v = kps.x, kps.y, kps.visible

    # calculate scaling for points and connections
    img_h, img_w, _ = img.shape
    img_area = img_h * img_w
    dynamic_size = int(0.01867599 * (img_area ** 0.4422045))
    dynamic_size = max(dynamic_size, 1)

    # draw connections
    if kps.metadata is not None and kps.metadata.connections is not None:
        for connection in kps.metadata.connections:
            if v[connection.p1] > 0 and v[connection.p2] > 0:
                cv2.line(
                    img,
                    (int(x[connection.p1]), int(y[connection.p1])),
                    (int(x[connection.p2]), int(y[connection.p2])),
                    color=connection.color,
                    thickness=dynamic_size,
                )

    # draw points
    for x_c, y_c in zip(x[v > 0], y[v > 0]):
        cv2.circle(
            img,
            (int(round(x_c)), int(round(y_c))),
            radius=dynamic_size,
            color=color,
            thickness=-1,
        )

    return img



__all__ = [
    "sort_losses",
    "get_stats",
    "get_weighted_sum",
    "add_annotations",
    "get_samples_losses",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.data import *
from icevision.visualize.show_data import show_preds


def get_weighted_sum(sample, weights):
    loss_weighted = 0
    for loss, weight in weights.items():
        loss_weighted += sample[loss] * weight
    sample["loss_weighted"] = loss_weighted
    return sample


def sort_losses(
    samples: List[dict], preds: List[dict], by: Union[str, dict] = "loss_total"
) -> Tuple[List[dict], List[dict], List[str]]:
    by_copy = deepcopy(by)
    losses_expected = [
        k for k in samples[0].keys() if "loss" in k and k != "loss_total"
    ]
    if "effdet_total_loss" in losses_expected:
        losses_expected.remove("effdet_total_loss")

    if isinstance(by, str):
        loss_check = losses_expected + ["loss_total"]
        assert (
            by in loss_check
        ), f"You must `sort_by` one of the losses. '{by}' is not among {loss_check}"

    if isinstance(by, dict):
        expected = ["weighted"]
        assert (
            by["method"] in expected
        ), f"`method` must be in {expected}, got {by['method']} instead."
        if by["method"] == "weighted":
            losses_passed = set(by["weights"].keys())
            losses_expected = set(losses_expected)
            assert (
                losses_passed == losses_expected
            ), f"You need to pass a weight for each of the losses in {losses_expected}, got {losses_passed} instead."

            samples = [get_weighted_sum(s, by["weights"]) for s in samples]
            by = "loss_weighted"

    l = list(zip(samples, preds))
    l = sorted(l, key=lambda i: i[0][by], reverse=True)
    sorted_samples, sorted_preds = zip(*l)
    annotations = [el["text"] for el in sorted_samples]

    if isinstance(by_copy, dict):
        if by_copy["method"] == "weighted":
            annotations = [
                f"loss_weighted: {round(s['loss_weighted'], 5)}\n" + a
                for a, s in zip(annotations, sorted_samples)
            ]

    return list(sorted_samples), list(sorted_preds), annotations


def get_stats(l: List) -> dict:
    l = np.array(l)
    quants_names = ["1ile", "25ile", "50ile", "75ile", "99ile"]
    quants = np.quantile(l, [0.01, 0.25, 0.5, 0.75, 0.99])
    d = {
        "min": l.min(),
        "max": l.max(),
        "mean": l.mean(),
    }

    q = {k: v for k, v in zip(quants_names, quants)}
    d.update(q)
    return d


def _move_to_device(x, y, device):

    if isinstance(y, list):
        x = [o.to(device) for o in x]
        y = [
            {
                k: (v.to(device) if isinstance(v, torch.Tensor) else v)
                for k, v in o.items()
            }
            for o in y
        ]
    elif isinstance(
        y, dict
    ):  # this covers the efficientdet case in which `y` is a dict of Union[list, tensor] and not a list of dicts and `x` is a Tensor and not a list of Tensors
        x = x.to(device)
        for k in y.keys():
            if isinstance(y[k], list):
                y[k] = [
                    o.to(device) if isinstance(o, torch.Tensor) else o for o in y[k]
                ]
            else:
                y[k] = y[k].to(device) if isinstance(y[k], torch.Tensor) else y[k]
    return x, y


def _prepend_str(d: dict, s: str):
    return {(s + "_" + k if s not in k else k): v for k, v in d.items()}


class Interpretation:
    def __init__(self, losses_dict, valid_dl, infer_dl, predict_dl):
        self.losses_dict = losses_dict
        self.valid_dl = valid_dl
        self.infer_dl = infer_dl
        self.predict_dl = predict_dl

    def _rename_losses(self, losses_dict):
        return losses_dict

    def _sum_losses(self, losses_dict):
        losses_dict["loss_total"] = sum(losses_dict.values())
        return losses_dict

    def get_losses(
        self,
        model: nn.Module,
        dataset: Dataset,
    ) -> Tuple[List[dict], dict]:
        """
        Gets a dataset and a model as input and returns losses calculated per image + losses stats.
        In case of faster_rcnn those are loss_classifier, loss_box_reg and loss_total
        (sum of the first 2 plus RPN related losses, e.g. sum of 4 terms).

        Arguments
        ---------
        model: nn.Module
        dataset: Dataset

        Returns
        -------
        samples_plus_losses: those are the samples (dictionaries) contained in the original dataset
                            enriched with losses.
        losses_stats: dictionary containing losses stats (min, max, mean, 1-25-50-75-99 quantiles)
                    for each one of the losses.
        """
        device = model_device(model)
        losses_stats = self.losses_dict
        dl = self.valid_dl(dataset, batch_size=1, num_workers=0, shuffle=False)

        samples_plus_losses = []

        with torch.no_grad():
            for (x, y), sample in pbar(dl):
                torch.manual_seed(0)
                x, y = _move_to_device(x, y, device)
                loss = model(x, y)
                loss = {k: float(v.cpu().numpy()) for k, v in loss.items()}
                loss = self._rename_losses(loss)
                loss = self._sum_losses(loss)

                for l in losses_stats.keys():
                    losses_stats[l].append(loss[l])

                loss = _prepend_str(loss, "loss")
                sample[0].update(loss)
                samples_plus_losses.append(sample[0])

        losses_stats = {k: get_stats(v) for k, v in losses_stats.items()}
        losses_stats = _prepend_str(losses_stats, "loss")
        return samples_plus_losses, losses_stats

    def plot_top_losses(
        self,
        model: nn.Module,
        dataset: Dataset,
        sort_by: str = "loss_total",
        n_samples: int = 5,
        batch_size: int = 8,
    ) -> Tuple[List[dict], List[dict], dict]:
        """
        Gets a dataset and a model as input. Calculates losses for each sample in the dataset.
        Sorts samples by `sort_by` (e.g. one of the losses). Runs model predictions on samples.
        Plots the top sorted `n_samples` and returns losses + predictions + losses_stats.

        Arguments
        ---------
        model: nn.Module
        dataset: Dataset
        sort_by: (str) the loss to sort samples by
        n_samples: how many samples to show
        batch_size: used when creating the infer dataloader to get model predictions on the dataset

        Returns
        -------
        sorted_samples: those are the samples (dictionaries) contained in the original dataset
                        enriched with losses and sorted by `sort_by`.
        sorted_preds: `model` predictions on `sorted_samples`. Same order as `sorted_samples`.
        losses_stats: dictionary containing losses stats (min, max, mean, 1-25-50-75-99 quantiles)
                    for each one of the losses.
        """
        logger.info(
            f"Losses returned by model: {[l for l in list(_prepend_str(self.losses_dict, 'loss').keys()) if l!='loss_total']}",
        )

        dl = self.valid_dl(dataset, batch_size=1, num_workers=0, shuffle=False)
        samples, losses_stats = self.get_losses(model, dataset)
        samples = add_annotations(samples)

        dl = self.infer_dl(dataset, batch_size=batch_size)
        _, preds = self.predict_dl(model=model, infer_dl=dl)

        sorted_samples, sorted_preds, annotations = sort_losses(
            samples, preds, by=sort_by
        )
        assert len(sorted_samples) == len(samples) == len(preds) == len(sorted_preds)

        anns = []
        for ann in annotations:
            ann = ann.split("\n")
            ann1 = "\n".join(ann[:4])
            ann2 = "\n".join(ann[4:])
            anns.append((ann1, ann2))

        show_preds(
            samples=sorted_samples[:n_samples],
            preds=sorted_preds[:n_samples],
            annotations=anns[:n_samples],
        )
        model.train()
        return sorted_samples, sorted_preds, losses_stats


def add_annotations(samples: List[dict]) -> List[dict]:
    """
    Adds a `text` field to the sample dict to use as annotations when plotting.
    """
    for sample in samples:
        text = ""
        for key in sample.keys():
            if "loss" in key:
                text += f"{key}: {round(sample[key], 5)}\n"
        text += f"IMG: {sample['filepath'].name}"
        sample["text"] = text
    return samples


def get_samples_losses(samples_plus_losses):
    return [
        {k: v for k, v in l.items() if "loss" in k or "file" in k}
        for l in samples_plus_losses
    ]



__all__ = ["ClassMap", "BACKGROUND"]

from icevision.imports import *

BACKGROUND = "background"


class ClassMap:
    """Utility class for mapping between class name and id."""

    def __init__(
        self,
        classes: Optional[Sequence[str]] = None,
        background: Optional[str] = BACKGROUND,
    ):
        self._lock = True

        self._id2class = copy(list(classes)) if classes else []
        # insert background if required
        self._background = background
        if self._background is not None:
            try:
                self._id2class.remove(self._background)
            except ValueError:
                pass
            # background is always index zero
            self._id2class.insert(0, self._background)

        self._class2id = {name: i for i, name in enumerate(self._id2class)}

    @property
    def num_classes(self):
        return len(self)

    def get_id(self, id: int) -> str:
        return self._id2class[id]

    def get_name(self, name: str) -> int:
        try:
            return self._class2id[name]
        except KeyError as e:
            if not self._lock:
                return self.add_name(name)
            else:
                raise e

    def add_name(self, name) -> int:
        self._id2class.append(name)
        id = len(self._class2id)
        self._class2id[name] = id
        return id

    def lock(self):
        self._lock = True

    def unlock(self):
        self._lock = False

    def __eq__(self, other) -> bool:
        if isinstance(other, ClassMap):
            return self.__dict__ == other.__dict__
        return False

    def __len__(self):
        return len(self._id2class)

    def __repr__(self):
        return f"<ClassMap: {self._class2id.__repr__()}>"



__all__ = [
    "RecordMixin",
    "ClassMapRecordMixin",
    "ImageidRecordMixin",
    "ImageRecordMixin",
    "FilepathRecordMixin",
    "SizeRecordMixin",
    "LabelsRecordMixin",
    "BBoxesRecordMixin",
    "MasksRecordMixin",
    "AreasRecordMixin",
    "IsCrowdsRecordMixin",
    "KeyPointsRecordMixin",
]

from icevision.imports import *
from icevision.utils import *
from .bbox import *
from .mask import *
from .exceptions import *
from .keypoints import *
from .class_map import *


class RecordMixin:
    # TODO: as_dict is only necessary because of backwards compatibility
    def as_dict(self) -> dict:
        return {}

    def _load(self) -> None:
        return

    def _num_annotations(self) -> Dict[str, int]:
        return {}

    def _autofix(self) -> Dict[str, bool]:
        return {}

    def _remove_annotation(self, i) -> None:
        return

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        return {}

    def _repr(self) -> List[str]:
        return []


class ClassMapRecordMixin(RecordMixin):
    def set_class_map(self, class_map: ClassMap):
        self.class_map = class_map

    def as_dict(self) -> dict:
        return {"class_map": self.class_map, **super().as_dict()}


class ImageidRecordMixin(RecordMixin):
    def set_imageid(self, imageid: int):
        self.imageid = imageid

    def _repr(self) -> List[str]:
        return [f"Image ID: {self.imageid}", *super()._repr()]

    def as_dict(self) -> dict:
        return {"imageid": self.imageid, **super().as_dict()}


# TODO: we need a way to combine filepath and image mixin
class ImageRecordMixin(RecordMixin):
    def set_img(self, img: np.ndarray):
        self.img = img
        self.height, self.width, _ = self.img.shape

    def _repr(self) -> List[str]:
        return [f"Image: {self.img}", *super()._repr()]

    def as_dict(self) -> dict:
        return {
            "img": self.img,
            "height": self.height,
            "width": self.width,
            **super().as_dict(),
        }


class FilepathRecordMixin(RecordMixin):
    def set_filepath(self, filepath: Union[str, Path]):
        self.filepath = Path(filepath)

    def _load(self):
        self.img = open_img(self.filepath)
        # TODO, HACK: is it correct to overwrite height and width here?
        self.height, self.width, _ = self.img.shape
        super()._load()

    def _autofix(self) -> Dict[str, bool]:
        exists = self.filepath.exists()
        if not exists:
            raise AutofixAbort(f"File '{self.filepath}' does not exist")

        return super()._autofix()

    def _repr(self) -> List[str]:
        return [f"Filepath: {self.filepath}", *super()._repr()]

    def as_dict(self) -> dict:
        # return {"filepath": self.filepath, **super().as_dict()}
        # HACK: img, height, width are conditonal, use __dict__ to circumvent that
        # can be resolved once `as_dict` gets removed
        return {**self.__dict__, **super().as_dict()}


class SizeRecordMixin(RecordMixin):
    def set_image_size(self, width: int, height: int):
        self.width, self.height = width, height

    def _repr(self) -> List[str]:
        return [
            f"Image size (width, height): ({self.width}, {self.height})",
            *super()._repr(),
        ]

    def as_dict(self) -> dict:
        return {"width": self.width, "height": self.height, **super().as_dict()}

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        info = [{"img_width": self.width, "img_height": self.height}]
        return {"img_size": info, **super()._aggregate_objects()}


### Annotation parsers ###
class LabelsRecordMixin(RecordMixin):
    def __init__(self):
        super().__init__()
        self.labels: List[int] = []

    def is_valid(self) -> List[bool]:
        return [True for _ in self.labels]

    def add_labels(self, labels: List[int]):
        self.labels.extend(labels)

    def _num_annotations(self) -> Dict[str, int]:
        return {"labels": len(self.labels), **super()._num_annotations()}

    def _autofix(self) -> Dict[str, bool]:
        return {"labels": [True] * len(self.labels), **super()._autofix()}

    def _remove_annotation(self, i):
        super()._remove_annotation(i)
        self.labels.pop(i)

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        return {"labels": self.labels, **super()._aggregate_objects()}

    def _repr(self) -> List[str]:
        return [f"Labels: {self.labels}", *super()._repr()]

    def as_dict(self) -> dict:
        return {"labels": self.labels, **super().as_dict()}


class BBoxesRecordMixin(RecordMixin):
    def __init__(self):
        super().__init__()
        self.bboxes: List[BBox] = []

    def _autofix(self) -> Dict[str, bool]:
        success = []
        for bbox in self.bboxes:
            try:
                autofixed = bbox.autofix(img_w=self.width, img_h=self.height)
                success.append(True)
            except InvalidDataError as e:
                logger.log("AUTOFIX-FAIL", "{}", str(e))
                success.append(False)

        return {"bboxes": success, **super()._autofix()}

    def add_bboxes(self, bboxes):
        self.bboxes.extend(bboxes)

    def _num_annotations(self) -> Dict[str, int]:
        return {"bboxes": len(self.bboxes), **super()._num_annotations()}

    def _remove_annotation(self, i):
        super()._remove_annotation(i)
        self.bboxes.pop(i)

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        objects = []
        for bbox in self.bboxes:
            x, y, w, h = bbox.xywh
            objects.append(
                {
                    "bbox_x": x,
                    "bbox_y": y,
                    "bbox_width": w,
                    "bbox_height": h,
                    "bbox_sqrt_area": bbox.area ** 0.5,
                    "bbox_aspect_ratio": w / h,
                }
            )

        return {"bboxes": objects, **super()._aggregate_objects()}

    def _repr(self) -> List[str]:
        return [f"BBoxes: {self.bboxes}", *super()._repr()]

    def as_dict(self) -> dict:
        return {"bboxes": self.bboxes, **super().as_dict()}


class MasksRecordMixin(RecordMixin):
    def __init__(self):
        super().__init__()
        self.masks = EncodedRLEs()

    def _load(self):
        super()._load()
        self.masks = MaskArray.from_masks(self.masks, self.height, self.width)

    def add_masks(self, masks: Sequence[Mask]):
        erles = [mask.to_erles(h=self.height, w=self.width) for mask in masks]
        self.masks.extend(erles)

    def _num_annotations(self) -> Dict[str, int]:
        return {"masks": len(self.masks), **super()._num_annotations()}

    def _remove_annotation(self, i):
        super()._remove_annotation(i)
        self.masks.pop(i)

    def _repr(self) -> List[str]:
        return [f"Masks: {self.masks}", *super()._repr()]

    def as_dict(self) -> dict:
        return {"masks": self.masks, **super().as_dict()}


class AreasRecordMixin(RecordMixin):
    def __init__(self):
        super().__init__()
        self.areas: List[float] = []

    def add_areas(self, areas: Sequence[float]):
        self.areas.extend(areas)

    def _num_annotations(self) -> Dict[str, int]:
        return {"areas": len(self.areas), **super()._num_annotations()}

    def _remove_annotation(self, i):
        super()._remove_annotation(i)
        self.areas.pop(i)

    def _repr(self) -> List[str]:
        return [f"Areas: {self.areas}", *super()._repr()]

    def as_dict(self) -> dict:
        return {"areas": self.areas, **super().as_dict()}


class IsCrowdsRecordMixin(RecordMixin):
    def __init__(self):
        super().__init__()
        self.iscrowds: List[bool] = []

    def add_iscrowds(self, iscrowds: Sequence[bool]):
        self.iscrowds.extend(iscrowds)

    def _num_annotations(self) -> Dict[str, int]:
        return {"iscrowds": len(self.iscrowds), **super()._num_annotations()}

    def _remove_annotation(self, i):
        super()._remove_annotation(i)
        self.iscrowds.pop(i)

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        return {"iscrowds": self.iscrowds, **super()._aggregate_objects()}

    def _repr(self) -> List[str]:
        return [f"Is Crowds: {self.iscrowds}", *super()._repr()]

    def as_dict(self) -> dict:
        return {"iscrowds": self.iscrowds, **super().as_dict()}


class KeyPointsRecordMixin(RecordMixin):
    def __init__(self):
        super().__init__()
        self.keypoints: List[KeyPoints] = []

    def add_keypoints(self, keypoints):
        self.keypoints.extend(keypoints)

    def _aggregate_objects(self) -> Dict[str, List[dict]]:
        objects = [
            {"keypoint_x": kpt.x, "keypoint_y": kpt.y, "keypoint_visible": kpt.v}
            for kpt in self.keypoints
        ]
        return {"keypoints": objects, **super()._aggregate_objects()}

    def _repr(self) -> List[str]:
        return {f"KeyPoints: {self.keypoints}", *super()._repr()}



__all__ = [
    "resnet18",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
    "resnext50_32x4d",
    "resnext101_32x8d",
    "wide_resnet50_2",
    "wide_resnet101_2",
]

from icevision.backbones.resnet import resnext101_32x8d
from icevision.backbones.resnet_fpn import (
    resnext50_32x4d,
    wide_resnet101_2,
    wide_resnet50_2,
)
from icevision.backbones import resnet_fpn

from torchvision.ops.feature_pyramid_network import LastLevelP6P7


def resnet18(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.resnet18(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )


def resnet34(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.resnet34(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )


def resnet50(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.resnet50(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )


def resnet101(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.resnet101(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )


def resnet152(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.resnet152(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )


def resnext50_32x4d(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.resnext50_32x4d(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )


def resnext101_32x8d(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.resnext101_32x8d(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )


def wide_resnet50_2(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.wide_resnet50_2(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )


def wide_resnet101_2(
    pretrained: bool = True,
    returned_layers=(2, 3, 4),
    extra_blocks=LastLevelP6P7(256, 256),
    **kwargs,
):
    return resnet_fpn.wide_resnet101_2(
        pretrained=pretrained,
        returned_layers=returned_layers,
        extra_blocks=extra_blocks,
        **kwargs,
    )



__all__ = [
    "wandb_img_preds",
]

from icevision.imports import *
from icevision.data import *
from icevision.core import *


def bbox_wandb(bbox: BBox, label: int, class_id_to_label, pred_score=None):
    """Creates a wandb compatible dictionary with bbox, label and score"""
    xmin, ymin, xmax, ymax = map(int, bbox.xyxy)

    box_data = {
        "position": {"minX": xmin, "maxX": xmax, "minY": ymin, "maxY": ymax},
        "class_id": label,
        "domain": "pixel",
    }

    if pred_score:
        score = int(pred_score * 100)
        box_caption = f"{class_id_to_label[label]} ({score}%)"
        box_data["score"] = score
    else:
        box_caption = f"{class_id_to_label[label]}"

    box_data["box_caption"] = box_caption

    return box_data


def wandb_image(sample, pred, class_id_to_label, add_ground_truth=False):
    raw_image = sample["img"]
    true_bboxes = sample["bboxes"]
    true_labels = sample["labels"]
    # Check if "masks" key is the sample dictionnary
    if "masks" in sample:
        true_masks = sample["masks"]

    pred_bboxes = pred["bboxes"]
    pred_labels = pred["labels"].tolist()
    pred_scores = pred["scores"]
    # Check if "masks" key is the pred dictionnary
    if "masks" in pred:
        pred_masks = pred["masks"]

    # Predicted Boxes
    pred_all_boxes = []
    # Collect predicted bounding boxes for this image
    for b_i, bbox in enumerate(pred_bboxes):
        box_data = bbox_wandb(
            bbox, pred_labels[b_i], class_id_to_label, pred_score=pred_scores[b_i]
        )
        pred_all_boxes.append(box_data)

    # log to wandb: raw image, predictions, and dictionary of class labels for each class id
    boxes = {
        "predictions": {"box_data": pred_all_boxes, "class_labels": class_id_to_label}
    }

    # Predicted Masks
    # Check if "masks" key is the pred dictionnary
    if "masks" in pred:
        mask_data = (pred_masks.data * pred["labels"][:, None, None]).max(0)
        masks = {
            "predictions": {"mask_data": mask_data, "class_labels": class_id_to_label}
        }
    else:
        masks = None

    # Ground Truth
    if add_ground_truth:
        # Ground Truth Boxes
        true_all_boxes = []
        # Collect ground truth bounding boxes for this image
        for b_i, bbox in enumerate(true_bboxes):
            box_data = bbox_wandb(bbox, true_labels[b_i], class_id_to_label)
            true_all_boxes.append(box_data)

        boxes["ground_truth"] = {
            "box_data": true_all_boxes,
            "class_labels": class_id_to_label,
        }

        # # Ground Truth Masks
        # Check if "masks" key is the sample dictionnary
        if "masks" in sample:
            labels_arr = np.array(sample["labels"])
            mask_data = (true_masks.data * labels_arr[:, None, None]).max(0)
            masks["ground_truth"] = {
                "mask_data": mask_data,
                "class_labels": class_id_to_label,
            }

    return wandb.Image(raw_image, boxes=boxes, masks=masks)


def wandb_img_preds(samples, preds, class_map, add_ground_truth=False):
    class_id_to_label = {int(v): k for k, v in class_map.class2id.items()}

    wandb_imgs = []
    for (sample, pred) in zip(samples, preds):
        img_wandb = wandb_image(
            sample, pred, class_id_to_label, add_ground_truth=add_ground_truth
        )
        wandb_imgs.append(img_wandb)
    return wandb_imgs



__all__ = ["InvalidDataError", "AutofixAbort", "AbortParseRecord"]


class InvalidDataError(Exception):
    pass


class AutofixAbort(Exception):
    pass


class AbortParseRecord(Exception):
    pass



__all__ = ["predict", "predict_dl", "convert_raw_prediction", "convert_raw_predictions"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.models.utils import _predict_dl


@torch.no_grad()
def predict(
    model: nn.Module,
    batch: Sequence[torch.Tensor],
    detection_threshold: float = 0.5,
    device: Optional[torch.device] = None,
):
    device = device or model_device(model)
    batch["img"] = [img.cuda() for img in batch["img"]]

    raw_pred = model(return_loss=False, rescale=False, **batch)
    return convert_raw_predictions(raw_pred, detection_threshold=detection_threshold)


def predict_dl(
    model: nn.Module, infer_dl: DataLoader, show_pbar: bool = True, **predict_kwargs
):
    return _predict_dl(
        predict_fn=predict,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        **predict_kwargs,
    )


def convert_raw_predictions(
    raw_preds: Sequence[Sequence[np.ndarray]], detection_threshold: float
):
    return [
        convert_raw_prediction(raw_pred, detection_threshold=detection_threshold)
        for raw_pred in raw_preds
    ]


def convert_raw_prediction(raw_pred: Sequence[np.ndarray], detection_threshold: float):
    scores, labels, bboxes = _unpack_raw_bboxes(raw_pred)

    keep_mask = scores > detection_threshold
    keep_scores = scores[keep_mask]
    keep_labels = labels[keep_mask]
    keep_bboxes = [BBox.from_xyxy(*o) for o in bboxes[keep_mask]]

    return {
        "scores": keep_scores,
        "labels": keep_labels,
        "bboxes": keep_bboxes,
    }


def _unpack_raw_bboxes(raw_bboxes):
    stack_raw_bboxes = np.vstack(raw_bboxes)

    scores = stack_raw_bboxes[:, -1]
    bboxes = stack_raw_bboxes[:, :-1]

    # each item in raw_pred is an array of predictions of it's `i` class
    labels = [np.full(o.shape[0], i, dtype=np.int32) for i, o in enumerate(raw_bboxes)]
    labels = np.concatenate(labels)

    return scores, labels, bboxes



__all__ = [
    "resnest50",
    "resnest101",
    "resnest200",
    "resnest269",
]

from icevision.imports import *
from icevision.utils import *
from icevision.backbones.resnet_fpn import patch_param_groups
from torchvision.ops.feature_pyramid_network import LastLevelMaxPool
from torchvision.models.detection.backbone_utils import (
    resnet_fpn_backbone,
    BackboneWithFPN,
)

from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.resnest:
    import resnest.torch


class Identity(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x


# TODO: resnest_name?
def resnest_fpn_backbone(
    backbone_fn,
    pretrained,
    # norm_layer=misc_nn_ops.FrozenBatchNorm2d,
    trainable_layers=3,
    returned_layers=None,
    extra_blocks=None,
):
    """
    Constructs a specified ResNest backbone with FPN on top. Freezes the specified number of layers in the backbone.

    Examples::

        >>> from torchvision.models.detection.backbone_utils import resnet_fpn_backbone
        >>> backbone = resnet_fpn_backbone('resnet50', pretrained=True, trainable_layers=3)
        >>> # get some dummy image
        >>> x = torch.rand(1,3,64,64)
        >>> # compute the output
        >>> output = backbone(x)
        >>> print([(k, v.shape) for k, v in output.items()])
        >>> # returns
        >>>   [('0', torch.Size([1, 256, 16, 16])),
        >>>    ('1', torch.Size([1, 256, 8, 8])),
        >>>    ('2', torch.Size([1, 256, 4, 4])),
        >>>    ('3', torch.Size([1, 256, 2, 2])),
        >>>    ('pool', torch.Size([1, 256, 1, 1]))]

    Arguments:
        backbone_name (string): resnet architecture. Possible values are 'ResNet', 'resnet18', 'resnet34', 'resnet50',
             'resnet101', 'resnet152', 'resnext50_32x4d', 'resnext101_32x8d', 'wide_resnet50_2', 'wide_resnet101_2'
        norm_layer (torchvision.ops): it is recommended to use the default value. For details visit:
            (https://github.com/facebookresearch/maskrcnn-benchmark/issues/267)
        pretrained (bool): If True, returns a model with backbone pre-trained on Imagenet
        trainable_layers (int): number of trainable (not frozen) resnet layers starting from final block.
            Valid values are between 0 and 5, with 5 meaning all backbone layers are trainable.
    """
    backbone = backbone_fn(pretrained=pretrained)

    backbone.fc = Identity()
    backbone.avgpool = Identity()

    # select layers that wont be frozen
    assert trainable_layers <= 5 and trainable_layers >= 0
    layers_to_train = ["layer4", "layer3", "layer2", "layer1", "conv1"][
        :trainable_layers
    ]
    # freeze layers only if pretrained backbone is used
    for name, parameter in backbone.named_parameters():
        if all([not name.startswith(layer) for layer in layers_to_train]):
            parameter.requires_grad_(False)

    if extra_blocks is None:
        extra_blocks = LastLevelMaxPool()

    if returned_layers is None:
        returned_layers = [1, 2, 3, 4]
    assert min(returned_layers) > 0 and max(returned_layers) < 5
    return_layers = {f"layer{k}": str(v) for v, k in enumerate(returned_layers)}

    in_channels_stage2 = backbone.inplanes // 8
    in_channels_list = [in_channels_stage2 * 2 ** (i - 1) for i in returned_layers]
    out_channels = 256
    return BackboneWithFPN(
        backbone,
        return_layers,
        in_channels_list,
        out_channels,
        extra_blocks=extra_blocks,
    )


def _resnest_fpn(name, pretrained: bool = True, **kwargs):
    model = resnest_fpn_backbone(backbone_fn=name, pretrained=pretrained, **kwargs)
    patch_param_groups(model)

    return model


def resnest50(pretrained: bool = True, **kwargs):
    return _resnest_fpn(resnest.torch.resnest50, pretrained=pretrained, **kwargs)


def resnest101(pretrained: bool = True, **kwargs):
    return _resnest_fpn(resnest.torch.resnest101, pretrained=pretrained, **kwargs)


def resnest200(pretrained: bool = True, **kwargs):
    return _resnest_fpn(resnest.torch.resnest200, pretrained=pretrained, **kwargs)


def resnest269(pretrained: bool = True, **kwargs):
    return _resnest_fpn(resnest.torch.resnest269, pretrained=pretrained, **kwargs)
